@@ -9,9 +9,12 @@
 #include <string.h>
 
 #include "auto/config.h"
+#include "nvim/api/private/defs.h"
+#include "nvim/api/private/helpers.h"
 #include "nvim/ascii_defs.h"
 #include "nvim/assert_defs.h"
 #include "nvim/charset.h"
+#include "nvim/errors.h"
 #include "nvim/eval/encode.h"
 #include "nvim/eval/typval.h"
 #include "nvim/eval/typval_defs.h"
@@ -19,12 +22,12 @@
 #include "nvim/garray.h"
 #include "nvim/garray_defs.h"
 #include "nvim/gettext_defs.h"
-#include "nvim/globals.h"
 #include "nvim/macros_defs.h"
 #include "nvim/math.h"
 #include "nvim/mbyte.h"
 #include "nvim/mbyte_defs.h"
 #include "nvim/memory.h"
+#include "nvim/memory_defs.h"
 #include "nvim/message.h"
 #include "nvim/option.h"
 #include "nvim/plines.h"
@@ -268,9 +271,9 @@ char *vim_strsave_shellescape(const char *string, bool do_special, bool do_newli
     }
     if (do_special && find_cmdline_var(p, &l) >= 0) {
       *d++ = '\\';                    // insert backslash
-      while (--l != SIZE_MAX) {  // copy the var
-        *d++ = *p++;
-      }
+      memcpy(d, p, l);                // copy the var
+      d += l;
+      p += l;
       continue;
     }
     if (*p == '\\' && fish_like) {
@@ -332,7 +335,7 @@ void vim_strcpy_up(char *restrict dst, const char *restrict src)
   while ((c = (uint8_t)(*src++)) != NUL) {
     *dst++ = (char)(uint8_t)(c < 'a' || c > 'z' ? c : c - 0x20);
   }
-  *dst = '\0';
+  *dst = NUL;
 }
 
 // strncpy (NUL-terminated) plus vim_strup.
@@ -343,7 +346,7 @@ void vim_strncpy_up(char *restrict dst, const char *restrict src, size_t n)
   while (n-- && (c = (uint8_t)(*src++)) != NUL) {
     *dst++ = (char)(uint8_t)(c < 'a' || c > 'z' ? c : c - 0x20);
   }
-  *dst = '\0';
+  *dst = NUL;
 }
 
 // memcpy (does not NUL-terminate) plus vim_strup.
@@ -472,6 +475,28 @@ bool striequal(const char *a, const char *b)
   return (a == NULL && b == NULL) || (a && b && STRICMP(a, b) == 0);
 }
 
+/// Compare two ASCII strings, for length "len", ignoring case, ignoring locale.
+///
+/// @return 0 for match, < 0 for smaller, > 0 for bigger
+int vim_strnicmp_asc(const char *s1, const char *s2, size_t len)
+  FUNC_ATTR_PURE FUNC_ATTR_WARN_UNUSED_RESULT
+{
+  int i = 0;
+  while (len > 0) {
+    i = TOLOWER_ASC(*s1) - TOLOWER_ASC(*s2);
+    if (i != 0) {
+      break;                       // this character is different
+    }
+    if (*s1 == NUL) {
+      break;                       // strings match until NUL
+    }
+    s1++;
+    s2++;
+    len--;
+  }
+  return i;
+}
+
 /// strchr() version which handles multibyte strings
 ///
 /// @param[in]  string  String to search in.
@@ -628,12 +653,14 @@ static const void *tv_ptr(const typval_T *const tvs, int *const idxp)
 #define OFF(attr) offsetof(union typval_vval_union, attr)
   STATIC_ASSERT(OFF(v_string) == OFF(v_list)
                 && OFF(v_string) == OFF(v_dict)
+                && OFF(v_string) == OFF(v_blob)
                 && OFF(v_string) == OFF(v_partial)
                 && sizeof(tvs[0].vval.v_string) == sizeof(tvs[0].vval.v_list)
                 && sizeof(tvs[0].vval.v_string) == sizeof(tvs[0].vval.v_dict)
+                && sizeof(tvs[0].vval.v_string) == sizeof(tvs[0].vval.v_blob)
                 && sizeof(tvs[0].vval.v_string) == sizeof(tvs[0].vval.v_partial),
-                "Strings, dictionaries, lists and partials are expected to be pointers, "
-                "so that all three of them can be accessed via v_string");
+                "Strings, Dictionaries, Lists, Blobs and Partials are expected to be pointers, "
+                "so that all of them can be accessed via v_string");
 #undef OFF
   const int idx = *idxp - 1;
   if (tvs[idx].v_type == VAR_UNKNOWN) {
@@ -793,10 +820,10 @@ static int format_typeof(const char *type)
   FUNC_ATTR_NONNULL_ALL
 {
   // allowed values: \0, h, l, L
-  char length_modifier = '\0';
+  char length_modifier = NUL;
 
   // current conversion specifier character
-  char fmt_spec = '\0';
+  char fmt_spec = NUL;
 
   // parse 'h', 'l', 'll' and 'z' length modifiers
   if (*type == 'h' || *type == 'l' || *type == 'z') {
@@ -864,7 +891,7 @@ static int format_typeof(const char *type)
     } else if (fmt_spec == 'd') {
       // signed
       switch (length_modifier) {
-      case '\0':
+      case NUL:
       case 'h':
         // char and short arguments are passed as int.
         return TYPE_INT;
@@ -878,7 +905,7 @@ static int format_typeof(const char *type)
     } else {
       // unsigned
       switch (length_modifier) {
-      case '\0':
+      case NUL:
       case 'h':
         return TYPE_UNSIGNEDINT;
       case 'l':
@@ -1002,7 +1029,7 @@ static void format_overflow_error(const char *pstart)
 
 enum { MAX_ALLOWED_STRING_WIDTH = 6400, };
 
-static int get_unsigned_int(const char *pstart, const char **p, unsigned *uj)
+static int get_unsigned_int(const char *pstart, const char **p, unsigned *uj, bool overflow_err)
 {
   *uj = (unsigned)(**p - '0');
   (*p)++;
@@ -1013,8 +1040,12 @@ static int get_unsigned_int(const char *pstart, const char **p, unsigned *uj)
   }
 
   if (*uj > MAX_ALLOWED_STRING_WIDTH) {
-    format_overflow_error(pstart);
-    return FAIL;
+    if (overflow_err) {
+      format_overflow_error(pstart);
+      return FAIL;
+    } else {
+      *uj = MAX_ALLOWED_STRING_WIDTH;
+    }
   }
 
   return OK;
@@ -1049,7 +1080,7 @@ static int parse_fmt_types(const char ***ap_types, int *num_posarg, const char *
       p += n;
     } else {
       // allowed values: \0, h, l, L
-      char length_modifier = '\0';
+      char length_modifier = NUL;
 
       // variable for positional arg
       int pos_arg = -1;
@@ -1075,7 +1106,7 @@ static int parse_fmt_types(const char ***ap_types, int *num_posarg, const char *
         // Positional argument
         unsigned uj;
 
-        if (get_unsigned_int(pstart, &p, &uj) == FAIL) {
+        if (get_unsigned_int(pstart, &p, &uj, tvs != NULL) == FAIL) {
           goto error;
         }
 
@@ -1118,7 +1149,7 @@ static int parse_fmt_types(const char ***ap_types, int *num_posarg, const char *
           // Positional argument field width
           unsigned uj;
 
-          if (get_unsigned_int(arg + 1, &p, &uj) == FAIL) {
+          if (get_unsigned_int(arg + 1, &p, &uj, tvs != NULL) == FAIL) {
             goto error;
           }
 
@@ -1144,7 +1175,7 @@ static int parse_fmt_types(const char ***ap_types, int *num_posarg, const char *
         const char *digstart = p;
         unsigned uj;
 
-        if (get_unsigned_int(digstart, &p, &uj) == FAIL) {
+        if (get_unsigned_int(digstart, &p, &uj, tvs != NULL) == FAIL) {
           goto error;
         }
 
@@ -1165,7 +1196,7 @@ static int parse_fmt_types(const char ***ap_types, int *num_posarg, const char *
             // Parse precision
             unsigned uj;
 
-            if (get_unsigned_int(arg + 1, &p, &uj) == FAIL) {
+            if (get_unsigned_int(arg + 1, &p, &uj, tvs != NULL) == FAIL) {
               goto error;
             }
 
@@ -1192,7 +1223,7 @@ static int parse_fmt_types(const char ***ap_types, int *num_posarg, const char *
           const char *digstart = p;
           unsigned uj;
 
-          if (get_unsigned_int(digstart, &p, &uj) == FAIL) {
+          if (get_unsigned_int(digstart, &p, &uj, tvs != NULL) == FAIL) {
             goto error;
           }
 
@@ -1440,7 +1471,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
       int space_for_positive = 1;
 
       // allowed values: \0, h, l, 2 (for ll), z, L
-      char length_modifier = '\0';
+      char length_modifier = NUL;
 
       // temporary buffer for simple numeric->string conversion
 #define TMP_LEN 350    // 1e308 seems reasonable as the maximum printable
@@ -1465,7 +1496,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
       size_t zero_padding_insertion_ind = 0;
 
       // current conversion specifier character
-      char fmt_spec = '\0';
+      char fmt_spec = NUL;
 
       // buffer for 's' and 'S' specs
       char *tofree = NULL;
@@ -1488,7 +1519,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
         const char *digstart = p;
         unsigned uj;
 
-        if (get_unsigned_int(digstart, &p, &uj) == FAIL) {
+        if (get_unsigned_int(digstart, &p, &uj, tvs != NULL) == FAIL) {
           goto error;
         }
 
@@ -1530,7 +1561,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
           // Positional argument field width
           unsigned uj;
 
-          if (get_unsigned_int(digstart, &p, &uj) == FAIL) {
+          if (get_unsigned_int(digstart, &p, &uj, tvs != NULL) == FAIL) {
             goto error;
           }
 
@@ -1539,15 +1570,19 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
           p++;
         }
 
-        const int j = (tvs
-                       ? (int)tv_nr(tvs, &arg_idx)
-                       : (skip_to_arg(ap_types, ap_start, &ap, &arg_idx,
-                                      &arg_cur, fmt),
-                          va_arg(ap, int)));
+        int j = (tvs
+                 ? (int)tv_nr(tvs, &arg_idx)
+                 : (skip_to_arg(ap_types, ap_start, &ap, &arg_idx,
+                                &arg_cur, fmt),
+                    va_arg(ap, int)));
 
         if (j > MAX_ALLOWED_STRING_WIDTH) {
-          format_overflow_error(digstart);
-          goto error;
+          if (tvs != NULL) {
+            format_overflow_error(digstart);
+            goto error;
+          } else {
+            j = MAX_ALLOWED_STRING_WIDTH;
+          }
         }
 
         if (j >= 0) {
@@ -1562,12 +1597,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
         const char *digstart = p;
         unsigned uj;
 
-        if (get_unsigned_int(digstart, &p, &uj) == FAIL) {
-          goto error;
-        }
-
-        if (uj > MAX_ALLOWED_STRING_WIDTH) {
-          format_overflow_error(digstart);
+        if (get_unsigned_int(digstart, &p, &uj, tvs != NULL) == FAIL) {
           goto error;
         }
 
@@ -1585,12 +1615,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
           const char *digstart = p;
           unsigned uj;
 
-          if (get_unsigned_int(digstart, &p, &uj) == FAIL) {
-            goto error;
-          }
-
-          if (uj > MAX_ALLOWED_STRING_WIDTH) {
-            format_overflow_error(digstart);
+          if (get_unsigned_int(digstart, &p, &uj, tvs != NULL) == FAIL) {
             goto error;
           }
 
@@ -1604,7 +1629,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
             // positional argument
             unsigned uj;
 
-            if (get_unsigned_int(digstart, &p, &uj) == FAIL) {
+            if (get_unsigned_int(digstart, &p, &uj, tvs != NULL) == FAIL) {
               goto error;
             }
 
@@ -1613,15 +1638,19 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
             p++;
           }
 
-          const int j = (tvs
-                         ? (int)tv_nr(tvs, &arg_idx)
-                         : (skip_to_arg(ap_types, ap_start, &ap, &arg_idx,
-                                        &arg_cur, fmt),
-                            va_arg(ap, int)));
+          int j = (tvs
+                   ? (int)tv_nr(tvs, &arg_idx)
+                   : (skip_to_arg(ap_types, ap_start, &ap, &arg_idx,
+                                  &arg_cur, fmt),
+                      va_arg(ap, int)));
 
           if (j > MAX_ALLOWED_STRING_WIDTH) {
-            format_overflow_error(digstart);
-            goto error;
+            if (tvs != NULL) {
+              format_overflow_error(digstart);
+              goto error;
+            } else {
+              j = MAX_ALLOWED_STRING_WIDTH;
+            }
           }
 
           if (j >= 0) {
@@ -1661,14 +1690,12 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
       }
 
       switch (fmt_spec) {
-      case 'b':
-      case 'B':
       case 'd':
       case 'u':
       case 'o':
       case 'x':
       case 'X':
-        if (tvs && length_modifier == '\0') {
+        if (tvs && length_modifier == NUL) {
           length_modifier = 'L';
         }
       }
@@ -1786,10 +1813,17 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
           if (ptr_arg) {
             arg_sign = 1;
           }
+        } else if (fmt_spec == 'b' || fmt_spec == 'B') {
+          uarg = (tvs
+                  ? (unsigned long long)tv_nr(tvs, &arg_idx)  // NOLINT(runtime/int)
+                  : (skip_to_arg(ap_types, ap_start, &ap, &arg_idx,
+                                 &arg_cur, fmt),
+                     va_arg(ap, unsigned long long)));  // NOLINT(runtime/int)
+          arg_sign = (uarg != 0);
         } else if (fmt_spec == 'd') {
           // signed
           switch (length_modifier) {
-          case '\0':
+          case NUL:
             arg = (tvs
                    ? (int)tv_nr(tvs, &arg_idx)
                    : (skip_to_arg(ap_types, ap_start, &ap, &arg_idx,
@@ -1835,7 +1869,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
         } else {
           // unsigned
           switch (length_modifier) {
-          case '\0':
+          case NUL:
             uarg = (tvs
                     ? (unsigned)tv_nr(tvs, &arg_idx)
                     : (skip_to_arg(ap_types, ap_start, &ap, &arg_idx,
@@ -2222,7 +2256,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap_st
   if (str_m > 0) {
     // make sure the string is nul-terminated even at the expense of
     // overwriting the last character (shouldn't happen, but just in case)
-    str[str_l <= str_m - 1 ? str_l : str_m - 1] = '\0';
+    str[str_l <= str_m - 1 ? str_l : str_m - 1] = NUL;
   }
 
   if (tvs != NULL
@@ -2819,10 +2853,10 @@ void f_strpart(typval_T *argvars, typval_T *rettv, EvalFuncData fptr)
   }
 
   if (argvars[2].v_type != VAR_UNKNOWN && argvars[3].v_type != VAR_UNKNOWN) {
-    int off;
+    int64_t off;
 
     // length in characters
-    for (off = (int)n; off < (int)slen && len > 0; len--) {
+    for (off = n; off < (int64_t)slen && len > 0; len--) {
       off += utfc_ptr2len(p + off);
     }
     len = off - n;
@@ -3128,3 +3162,39 @@ void f_trim(typval_T *argvars, typval_T *rettv, EvalFuncData fptr)
   }
   rettv->vval.v_string = xstrnsave(head, (size_t)(tail - head));
 }
+
+/// compare two keyvalue_T structs by case sensitive value
+int cmp_keyvalue_value(const void *a, const void *b)
+{
+  keyvalue_T *kv1 = (keyvalue_T *)a;
+  keyvalue_T *kv2 = (keyvalue_T *)b;
+
+  return strcmp(kv1->value, kv2->value);
+}
+
+/// compare two keyvalue_T structs by value with length
+int cmp_keyvalue_value_n(const void *a, const void *b)
+{
+  keyvalue_T *kv1 = (keyvalue_T *)a;
+  keyvalue_T *kv2 = (keyvalue_T *)b;
+
+  return strncmp(kv1->value, kv2->value, MAX(kv1->length, kv2->length));
+}
+
+/// compare two keyvalue_T structs by case insensitive value
+int cmp_keyvalue_value_i(const void *a, const void *b)
+{
+  keyvalue_T *kv1 = (keyvalue_T *)a;
+  keyvalue_T *kv2 = (keyvalue_T *)b;
+
+  return STRICMP(kv1->value, kv2->value);
+}
+
+/// compare two keyvalue_T structs by case insensitive value with length
+int cmp_keyvalue_value_ni(const void *a, const void *b)
+{
+  keyvalue_T *kv1 = (keyvalue_T *)a;
+  keyvalue_T *kv2 = (keyvalue_T *)b;
+
+  return STRNICMP(kv1->value, kv2->value, MAX(kv1->length, kv2->length));
+}
