@@ -16,7 +16,6 @@ describe('search highlighting', function()
   before_each(function()
     clear()
     screen = Screen.new(40, 7)
-    screen:attach()
     screen:set_default_attr_ids({
       [1] = { bold = true, foreground = Screen.colors.Blue },
       [2] = { background = Screen.colors.Yellow }, -- Search
@@ -74,10 +73,10 @@ describe('search highlighting', function()
     -- 'hlsearch' is enabled by default. #2859
     feed('gg/text<cr>')
     screen:expect([[
-        some {2:^text}                             |
-        more {2:text}stuff                        |
-        stupid{2:texttext}stuff                   |
-        a {2:text} word                           |
+      some {2:^text}                               |
+      more {2:text}stuff                          |
+      stupid{2:texttext}stuff                     |
+      a {2:text} word                             |
                                               |
       {1:~                                       }|
       /text                                   |
@@ -86,10 +85,10 @@ describe('search highlighting', function()
     -- overlapping matches not allowed
     feed('3nx')
     screen:expect([[
-        some {2:text}                             |
-        more {2:text}stuff                        |
-        stupid{2:text}^extstuff                    |
-        a {2:text} word                           |
+      some {2:text}                               |
+      more {2:text}stuff                          |
+      stupid{2:text}^extstuff                      |
+      a {2:text} word                             |
                                               |
       {1:~                                       }|
       /text                                   |
@@ -97,10 +96,10 @@ describe('search highlighting', function()
 
     feed('ggn*') -- search for entire word
     screen:expect([[
-        some {2:text}                             |
-        more textstuff                        |
-        stupidtextextstuff                    |
-        a {2:^text} word                           |
+      some {2:text}                               |
+      more textstuff                          |
+      stupidtextextstuff                      |
+      a {2:^text} word                             |
                                               |
       {1:~                                       }|
       /\<text\>                               |
@@ -108,10 +107,10 @@ describe('search highlighting', function()
 
     feed_command('nohlsearch')
     screen:expect([[
-        some text                             |
-        more textstuff                        |
-        stupidtextextstuff                    |
-        a ^text word                           |
+      some text                               |
+      more textstuff                          |
+      stupidtextextstuff                      |
+      a ^text word                             |
                                               |
       {1:~                                       }|
       :nohlsearch                             |
@@ -197,7 +196,8 @@ describe('search highlighting', function()
       }
     end)
 
-    it('works for multiline match', function()
+    -- oldtest: Test_hlsearch_cursearch()
+    it('works for multiline match, no duplicate highlight', function()
       command([[call setline(1, ['one', 'foo', 'bar', 'baz', 'foo the foo and foo', 'bar'])]])
       feed('gg/foo<CR>')
       screen:expect([[
@@ -281,6 +281,28 @@ describe('search highlighting', function()
         {2:hij}kl                                   |
         /efg\nhij                               |
       ]])
+
+      -- check clearing CurSearch when using it for another match
+      feed('G?^abcd<CR>Y')
+      screen:expect([[
+        ---                                     |
+        {1:abcd}efg                                 |
+        hijkl                                   |
+        ---                                     |
+        {2:^abcd}efg                                 |
+        hijkl                                   |
+        ?^abcd                                  |
+      ]])
+      feed('kkP')
+      screen:expect([[
+        ---                                     |
+        {1:abcd}efg                                 |
+        {2:^abcd}efg                                 |
+        hijkl                                   |
+        ---                                     |
+        {1:abcd}efg                                 |
+        ?^abcd                                  |
+      ]])
     end)
   end)
 
@@ -345,12 +367,19 @@ describe('search highlighting', function()
       bar baz foo
       bar foo baz]])
     feed('/foo')
+    screen:set_default_attr_ids({
+      [1] = { bold = true, foreground = Screen.colors.Blue },
+      [2] = { background = Screen.colors.Yellow }, -- Search
+      [3] = { reverse = true },
+      [4] = { bold = true, reverse = true },
+      [5] = { foreground = Screen.colors.White, background = Screen.colors.DarkGreen },
+    })
     screen:expect([[
       {3:foo} bar baz         │{MATCH:%d+}: {2:foo}{MATCH:%s+}|
       bar baz {2:foo}         │{MATCH:%d+}: {2:foo}{MATCH:%s+}|
       bar {2:foo} baz         │{MATCH:%d+}: {2:foo}{MATCH:%s+}|
       {1:~                   }│{MATCH:.*}|*2
-      {5:[No Name] [+]        }{3:term               }|
+      {4:[No Name] [+]        }{5:term               }|
       /foo^                                    |
     ]])
   end)
@@ -612,7 +641,7 @@ describe('search highlighting', function()
     feed_command('/ial te')
     screen:expect {
       grid = [[
-        very {5:spec^ial}{2: te}{6:xt}                     |
+      very {5:spec^ial}{2: te}{6:xt}                       |
                                               |
       {1:~                                       }|*4
       {4:search hit BOTTOM, continuing at TOP}    |
@@ -623,7 +652,7 @@ describe('search highlighting', function()
           topline = 0,
           botline = 3,
           curline = 0,
-          curcol = 11,
+          curcol = 9,
           linecount = 2,
           sum_scroll_delta = 0,
         },
@@ -641,7 +670,7 @@ describe('search highlighting', function()
     }
     command('%foldopen')
     screen:expect([[
-        very {5:spec^ial}{2: te}{6:xt}                     |
+      very {5:spec^ial}{2: te}{6:xt}                       |
                                               |
       {1:~                                       }|*4
       {4:search hit BOTTOM, continuing at TOP}    |
@@ -649,7 +678,7 @@ describe('search highlighting', function()
 
     feed_command('call clearmatches()')
     screen:expect([[
-        very spec{2:^ial te}xt                     |
+      very spec{2:^ial te}xt                       |
                                               |
       {1:~                                       }|*4
       :call clearmatches()                    |
@@ -659,7 +688,7 @@ describe('search highlighting', function()
     -- nonconflicting attributes are combined
     feed_command('syntax keyword MyGroup special')
     screen:expect([[
-        very {5:spec}{7:^ial}{2: te}xt                     |
+      very {5:spec}{7:^ial}{2: te}xt                       |
                                               |
       {1:~                                       }|*4
       :syntax keyword MyGroup special         |
